/*
 * Copyright (C) 1993,
 *
 * Department of Computing Science,
 * The University,
 * Newcastle upon Tyne,
 * UK.
 *
 * $Id: Resource.cc,v 1.1 1995/02/28 13:26:10 nmcl Exp $
 */

#include <iostream>

#ifndef RESOURCE_H_
#   include "Resource.h"
#endif

Boolean Resources::newCalled = FALSE;


Resources::Resources ()
                   : _referenceCount(0),
                     heapAllocated(newCalled)
{
    Resources::newCalled = FALSE;
}

Resources::~Resources ()
{
    if ((_referenceCount != 0) && (heapAllocated))
        std::cerr << "Resource::~Resource : _referenceCount = "
	     << _referenceCount << std::endl;
}

/*
 * If allocated on the stack then we could simply ignore all
 * ref and unref requests. Currently we process them but
 * treat "error" conditions differently.
 */

void Resources::ref (Resources* resource)
{
    if (resource != NULL)
        resource->_referenceCount++;
}

Boolean Resources::unref (Resources* resource)
{
    if (resource != (Resources*) 0)
    {
        if (resource->_referenceCount > 0)
            resource->_referenceCount--;
        else
	    if (resource->heapAllocated)
	        std::cerr << "Resource::unref : _referenceCount == 0" << std::endl;

        if ((resource->_referenceCount == 0) && (resource->heapAllocated))
        {
            delete resource;
            resource = (Resources*) 0;
        }
    }

    if ((resource != (Resources*) 0) && (resource->heapAllocated))
        return FALSE;
    else
        return TRUE;
}

void* Resources::operator new (size_t amount)
{
    Resources::newCalled = TRUE;
    return ::new char[amount];
}

void Resources::operator delete (void* memPtr)
{
    ::delete [] memPtr;
}
